
# Copyright (c) 2021 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

from collections import namedtuple

import pytest

from todo_txt_gtd.utils import is_task

Case = namedtuple("Case", ["line", "is_task"])


@pytest.mark.parametrize(
    "case",
    [
        Case("# foo @context", False),
        Case(" # foo @context", False),
        Case("", False),
        Case("x foo @context", False),
        Case("foo", False),
        Case("bar @context", False),
        Case("foo @context", True),
    ]
)
def test_utils_is_task(case):
    assert is_task(case.line, "foo") == case.is_task



#!/usr/bin/python3

import os
import shlex
import sys
import subprocess

infile = sys.argv[1]
outfile = infile[:-3] + ".tasks"

subprocess.call(shlex.split("../../tdtlist -f {}".format(infile)))

os.remove("tasks.odt")
os.rename("tasks.txt", outfile)




# Copyright (c) 2021 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import os
import re
from contextlib import contextmanager
from functools import wraps


def none_on_exception(*exceptions):
    def _none_on_exception(fp):
        @wraps(fp)
        def wrapper(*args, **kwargs):
            try:
                return fp(*args, **kwargs)
            except exceptions:
                return None

        return wrapper

    return _none_on_exception


@contextmanager
def nullfd(fd):
    saveout = os.dup(fd)
    os.close(fd)
    os.open(os.devnull, os.O_RDWR)
    try:
        yield
    finally:
        os.dup2(saveout, fd)
        os.close(saveout)


def is_task(line: str, *terms: str) -> bool:
    if re.search(r"^\s*#", line):
        return False

    if "@" not in line:
        return False

    if re.search("^x ", line):
        return False

    if any(term not in line for term in terms):
        return False

    return True


def is_current_task(line: str, *terms: str) -> bool:
    if "@~" in line:
        return False

    if threshold_mask(line):
        return False

    return is_task(line, *terms)


def threshold_mask(task: str) -> bool:
    match = re.search(r"(^|[^\S])t:(\d\d\d\d-\d\d-\d\d)($|[^\S])", task)

    if not match:
        return False

    threshold_date = datetime.datetime.strptime(match.group(2), "%Y-%m-%d")

    return datetime.datetime.now() < threshold_date



#!/usr/bin/python3
# Copyright (c) 2021 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import argparse
import copy
import os
import re
from collections import OrderedDict

from .utils import none_on_exception

NONE_PROJ = "_None"


@none_on_exception(AttributeError)
def TaskProj(line):
    return re.search(r" \+([^ ]+)", line).group(1)  # noqa


@none_on_exception(AttributeError)
def HeaderProj(line, prev_line):
    if prev_line.startswith("#"):
        return None

    return re.search("^# ([^ ]+)$", line).group(1)  # noqa


class Projects(OrderedDict):
    def __init__(self, todotxt=""):
        super(Projects, self).__init__()
        self[NONE_PROJ]

        for proj_name, task_text in self.LineGen(todotxt):
            self[proj_name].AddTask(task_text)

    def __missing__(self, key):
        if key is None:
            proj = self[NONE_PROJ]
        else:
            proj = Project(key)
            self[key] = proj
        return proj

    def __str__(self):
        projs = sorted((str(self[x]) for x in self), key=lambda y: y.upper())
        return "\n".join(projs)

    def LineGen(self, todotxt):
        current_project = None

        deferredTasks = []
        prev_line = ""

        for line in todotxt.split("\n"):

            if HeaderProj(line, prev_line):
                current_project = HeaderProj(line, prev_line)

            taskProj = TaskProj(line)
            if taskProj:
                if current_project and (current_project != taskProj):
                    deferredTasks.append(line)
                else:
                    yield (taskProj, line)
            else:
                if current_project:
                    yield (current_project, line)
                else:
                    deferredTasks.append(line)

            prev_line = line

        for line in deferredTasks:
            yield (TaskProj(line), line)


class Project(object):
    def __init__(self, name):
        self.name = name
        self.tasks = []

    def __len__(self):
        return len(self.tasks)

    def __iter__(self):
        return iter(self.tasks)

    def __copy__(self):
        image = Project(self.name)
        image.tasks = copy.copy(self.tasks)
        return image

    def AddTask(self, text):
        if not self.tasks and (not text or text[0] != "#"):
            self.AddTask("# {}".format(self.name))
            self.AddTask("#")
            self.AddTask("")

        if not (self.tasks and HeaderProj(text, self.tasks[-1].text)):
            self.tasks.append(Task(text, self.name))

    def finish(self):
        if not self.tasks or self.tasks[-1].text:
            self.AddTask("")
        return self

    def __str__(self):
        return "\n".join(str(x) for x in copy.copy(self).finish())


class Task(object):
    def __init__(self, text, project):
        self.text = self.FixTask(text, project)

    @none_on_exception(AttributeError)
    def GetContext(self, text):
        return re.search(" @([^ ]+)", text).group(1)

    @none_on_exception(AttributeError)
    def GetProject(self, text):
        return re.search(r" \+([^ ]+)", text).group(1)  # noqa

    def FixTask(self, text, project):
        if (
            self.GetContext(text) is not None
            and self.GetProject(text) is None
            and project != NONE_PROJ
            and project is not None
        ):

            text = "{0} +{1}".format(text, project)

        return text

    def __str__(self):
        return self.text


def parse_args():
    parser = argparse.ArgumentParser(
        description="Clean up the todo.txt file in a GTD fashion"
    )

    parser.add_argument(
        "file",
        help="the todo.txt file location ",
    )

    args = parser.parse_args()
    args.file = os.path.expanduser(args.file)

    return args


def cleanup(filepath):
    with open(filepath, "r", encoding="utf-8") as fp:
        projects = Projects(fp.read())

    with open(filepath, "w", encoding="utf-8") as fp:
        fp.write(str(projects))


def main():
    args = parse_args()

    cleanup(args.file)


if __name__ == "__main__":
    main()




# Copyright (c) 2021 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE


import argparse
import os
import re
import subprocess
import textwrap
from tempfile import TemporaryDirectory

from .tdtcleanup import Project, Projects
from .utils import is_task


def default_file():
    cp = subprocess.run(
        ["todo.txt", "--info"],
        capture_output=True,
        encoding="utf-8",
    )

    match = re.search("^task_path\s*=\s*(.+)$", cp.stdout, re.MULTILINE)
    return match.group(1)


def parse_args():
    parser = argparse.ArgumentParser(
        description="Work with one or more GTD projects in todo.txt",
        epilog=textwrap.dedent(
            """
            Edit one or more isolated projects in a todo.txt file
            (todo.txt projects are denoted by a a leading "+").

            If the entire project is deleted during the edit session,
            the original project is preserved in todo.txt. If just the
            Project Header line is kept, then the project is deleted in
            the original.

            The default text editor, set by 'update-alternatives', is
            used. This can be overridden by setting the 'EDITOR' environment
            variable.
        """[
                1:-1
            ]
        ),
    )

    parser.add_argument(
        "-f",
        "--file",
        help="the todo.txt file location.",
        default=None,
    )

    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="just list the projects in the current todo.txt file",
    )

    parser.add_argument(
        "-x",
        "--exact",
        action="store_true",
        help="require an exact match of project to TERM",
    )

    parser.add_argument(
        "terms",
        nargs="*",
        metavar="TERM",
        help="search terms to filter the project(s) to use. "
        "Projects matching ANY of the terms will be used.",
    )

    args = parser.parse_args()

    if args.file is None:
        args.file = default_file()

    args.file = os.path.expanduser(args.file)

    return args


def read_proj(path):
    with open(path, "r", encoding="utf-8") as p:
        return Projects(p.read())


def write_proj(path, projs):
    with open(path, "w", encoding="utf-8") as p:
        p.write(str(projs))


def save_selected_projs(tdpath, editpath, terms, exact):
    pdict = read_proj(tdpath)

    for project in list(pdict):
        if terms and not any(x in project for x in terms):
            del pdict[project]
        elif exact and terms and not any(x == project for x in terms):
            del pdict[project]

    if len(pdict) == 0 and terms and terms[0]:
        proj = Project(terms[0])
        pdict[terms[0]] = proj

    write_proj(editpath, pdict)

    return {x for x in pdict if x != "_None"}


def edit_proj(tdpath, terms, exact):
    with TemporaryDirectory() as tmpdir:
        editpath = os.path.join(tmpdir, "todo.txt")
        projhdrs = save_selected_projs(tdpath, editpath, terms, exact)

        try:
            editor = os.environ["EDITOR"]
        except KeyError:
            editor = "editor"
        subprocess.run([editor, editpath])

        editprojs = read_proj(editpath)

    allprojs = read_proj(tdpath)

    for proj in editprojs:
        if len(editprojs[proj]) <= 2 and proj != "_None":
            if proj in allprojs:
                del allprojs[proj]
        elif proj in projhdrs:
            allprojs[proj] = editprojs[proj]
        else:
            for task in editprojs[proj]:
                if is_task(str(task)):
                    allprojs[proj].AddTask(str(task))

    write_proj(tdpath, allprojs)


def main():
    args = parse_args()

    if args.list:
        for proj in read_proj(args.file):
            if not args.terms:
                print(proj)
            elif args.exact and any(x == proj for x in args.terms):
                print(proj)
            elif not args.exact and any(x in proj for x in args.terms):
                print(proj)
    else:
        edit_proj(args.file, args.terms, args.exact)


if __name__ == "__main__":
    main()




# Copyright (c) 2021 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import copy

import pytest

from todo_txt_gtd import tdtcleanup


@pytest.mark.parametrize("numruns", [1, 2])
def test_cleanup(clean_fxt, numruns):
    for _ in range(numruns):
        tdtcleanup.cleanup(str(clean_fxt.workfile))

    test_output = clean_fxt.workfile.read_text("utf-8")
    ref_output = clean_fxt.outfile.read_text("utf-8")
    assert test_output == ref_output
    assert clean_fxt.workfile.size() == clean_fxt.outfile.size()


@pytest.fixture
def projs_fxt(file_case):
    with open(str(file_case.outfile), "r") as fp:
        text = fp.read()

    return tdtcleanup.Projects(text)


def test_proj_copy(projs_fxt):
    for proj in projs_fxt:
        assert str(proj) == str(copy.copy(proj))


def test_null_proj_str():
    proj = tdtcleanup.Project("foo")
    assert "foo" in str(proj)

    assert proj.tasks == []




# Copyright (c) 2021 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import os
from collections import namedtuple
import py
import pytest


def cases():
    Case = namedtuple("Case", ["infile", "outfile", "taskfile"])

    testdir = os.path.split(__file__)[0]
    casepath = os.path.join(testdir, "cases/")

    cases = []
    for path in os.listdir(casepath):
        if path[-3:] == ".in":
            cases.append(
                Case(
                    py.path.local(casepath + path),
                    py.path.local(casepath + path[:-3] + ".out"),
                    py.path.local(casepath + path[:-3] + ".tasks"),
                )
            )

    return cases


def pytest_generate_tests(metafunc):
    if "file_case" in metafunc.fixturenames:
        metafunc.parametrize("file_case", cases())


def makefiles(file_case, tmpdir):
    workfile = tmpdir.join("todo.txt")
    file_case.infile.copy(workfile)

    outfile = tmpdir.join(file_case.outfile.basename)
    file_case.outfile.copy(outfile)

    taskfile = tmpdir.join(file_case.taskfile.basename)
    file_case.taskfile.copy(taskfile)

    return (outfile, workfile, taskfile)


@pytest.fixture
def clean_fxt(file_case, tmpdir):

    Clean = namedtuple("Clean", ["outfile", "workfile", "taskfile"])
    return Clean(*makefiles(file_case, tmpdir))



#!/usr/bin/python3
# Copyright (c) 2021 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

from os import path

from setuptools import setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="todo.txt-gtd",
    version="0.2",
    packages=["todo_txt_gtd"],
    url="https://github.com/davesteele/todo.txt-gtd",
    license="GPL 2.0",
    author="David Steele",
    author_email="dsteele@gmail.com",
    description="Todo.txt support scripts for GTD.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": [
            "gtdcleanup = todo_txt_gtd.tdtcleanup:main",
            "project = todo_txt_gtd.tdtproj:main",
        ]
    },
    install_requires=["configargparse", "relatorio"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: "
        "GNU General Public License v2 or later (GPLv2+)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Other/Nonlisted Topic",
    ],
)
